@@ -45,10 +45,6 @@ int le_ssh2_listener;
 int le_ssh2_sftp;
 int le_ssh2_pkey_subsys;
 
-ZEND_BEGIN_ARG_INFO(php_ssh2_first_arg_force_ref, 0)
-	ZEND_ARG_PASS_INFO(1)
-ZEND_END_ARG_INFO()
-
 /* *************
    * Callbacks *
    ************* */
@@ -100,8 +96,13 @@ LIBSSH2_DEBUG_FUNC(php_ssh2_debug_cb)
 	ZVAL_STRINGL(&args[1], language, language_len);
 	ZVAL_LONG(&args[2], always_display);
 
-	if (FAILURE == call_user_function_ex(NULL, NULL, data->disconnect_cb, NULL, 3, args, 0, NULL)) {
-		php_error_docref(NULL, E_WARNING, "Failure calling disconnect callback");
+	zval retval;
+	if (FAILURE == call_user_function(NULL, NULL, data->debug_cb, &retval, 3, args)) {
+		php_error_docref(NULL, E_WARNING, "Failure calling debug callback");
+	}
+
+	if (!Z_ISUNDEF(retval)) {
+		zval_ptr_dtor(&retval);
 	}
 }
 /* }}} */
@@ -125,7 +126,7 @@ LIBSSH2_IGNORE_FUNC(php_ssh2_ignore_cb)
 
 	ZVAL_STRINGL(&args[0], message, message_len);
 
-	if (FAILURE == call_user_function_ex(NULL, NULL, data->ignore_cb, &zretval, 1, args, 0, NULL)) {
+	if (FAILURE == call_user_function(NULL, NULL, data->ignore_cb, &zretval, 1, args)) {
 		php_error_docref(NULL, E_WARNING, "Failure calling ignore callback");
 	}
 	if (Z_TYPE_P(&zretval) != IS_UNDEF) {
@@ -155,7 +156,7 @@ LIBSSH2_MACERROR_FUNC(php_ssh2_macerror_cb)
 
 	ZVAL_STRINGL(&args[0], packet, packet_len);
 
-	if (FAILURE == call_user_function_ex(NULL, NULL, data->macerror_cb, &zretval, 1, args, 0, NULL)) {
+	if (FAILURE == call_user_function(NULL, NULL, data->macerror_cb, &zretval, 1, args)) {
 		php_error_docref(NULL, E_WARNING, "Failure calling macerror callback");
 	} else {
 		retval = zval_is_true(&zretval) ? 0 : -1;
@@ -188,9 +189,14 @@ LIBSSH2_DISCONNECT_FUNC(php_ssh2_disconnect_cb)
 	ZVAL_STRINGL(&args[1], message, message_len);
 	ZVAL_STRINGL(&args[2], language, language_len);
 
-	if (FAILURE == call_user_function_ex(NULL, NULL, data->disconnect_cb, NULL, 3, args, 0, NULL)) {
+	zval retval;
+	if (FAILURE == call_user_function(NULL, NULL, data->disconnect_cb, &retval, 3, args)) {
 		php_error_docref(NULL, E_WARNING, "Failure calling disconnect callback");
 	}
+
+	if (!Z_ISUNDEF(retval)) {
+		zval_ptr_dtor(&retval);
+	}
 }
 /* }}} */
 
@@ -220,8 +226,8 @@ static int php_ssh2_set_callback(LIBSSH2_SESSION *session, HashTable *ht, char *
 		return -1;
 	}
 
-	copyval = handler;
-	zval_copy_ctor(copyval);
+	copyval = emalloc(sizeof(zval));
+	ZVAL_COPY(copyval, handler);
 
 	switch (callback_type) {
 		case LIBSSH2_CALLBACK_IGNORE:
@@ -440,9 +446,7 @@ PHP_FUNCTION(ssh2_disconnect)
 		RETURN_FALSE;
 	}
 
-	if (zend_list_close(Z_RES_P(zsession)) != SUCCESS) {
-		RETURN_FALSE;
-	}
+	zend_list_close(Z_RES_P(zsession));
 
 	RETURN_TRUE;
 }
@@ -649,21 +653,17 @@ PHP_FUNCTION(ssh2_auth_pubkey_file)
 {
 	LIBSSH2_SESSION *session;
 	zval *zsession;
-	char *username, *pubkey, *privkey, *passphrase = NULL;
-	size_t username_len, pubkey_len, privkey_len, passphrase_len = 0;
+	zend_string *username, *pubkey, *privkey, *passphrase;
 #ifndef PHP_WIN32
-	char *newpath;
+	zend_string *newpath;
 	struct passwd *pws;
 #endif
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), "rsss|s", &zsession,	&username, &username_len,
-																				&pubkey, &pubkey_len,
-																				&privkey, &privkey_len,
-																				&passphrase, &passphrase_len) == FAILURE) {
+	if (zend_parse_parameters(ZEND_NUM_ARGS(), "rSSS|S", &zsession,	&username, &pubkey, &privkey, &passphrase) == FAILURE) {
 		return;
 	}
 
-	if (php_check_open_basedir(pubkey) || php_check_open_basedir(privkey)) {
+	if (php_check_open_basedir(ZSTR_VAL(pubkey)) || php_check_open_basedir(ZSTR_VAL(privkey))) {
 		RETURN_FALSE;
 	}
 
@@ -672,28 +672,28 @@ PHP_FUNCTION(ssh2_auth_pubkey_file)
 	/* Explode '~/paths' stopgap fix because libssh2 does not accept tilde for homedir
 	  This should be ifdef'ed when a fix is available to support older libssh2 versions*/
 	pws = getpwuid(geteuid());
-	if (pubkey_len >= 2 && *pubkey == '~' && *(pubkey+1) == '/') {
-		newpath = emalloc(strlen(pws->pw_dir) + strlen(pubkey));
-		strcpy(newpath, pws->pw_dir);
-		strcat(newpath, pubkey+1);
-		efree(pubkey);
+	if (ZSTR_LEN(pubkey) >= 2 && *(ZSTR_VAL(pubkey)) == '~' && *(ZSTR_VAL(pubkey)+1) == '/') {
+		newpath = zend_string_alloc(strlen(pws->pw_dir) + ZSTR_LEN(pubkey), 0);
+		strcpy(ZSTR_VAL(newpath), pws->pw_dir);
+		strcat(ZSTR_VAL(newpath), ZSTR_VAL(pubkey)+1);
+		zend_string_release(pubkey);
 		pubkey = newpath;
 	}
-	if (privkey_len >= 2 && *privkey == '~' && *(privkey+1) == '/') {
-		newpath = emalloc(strlen(pws->pw_dir) + strlen(privkey));
-		strcpy(newpath, pws->pw_dir);
-		strcat(newpath, privkey+1);
-		efree(privkey);
+	if (ZSTR_LEN(privkey) >= 2 && *(ZSTR_VAL(privkey)) == '~' && *(ZSTR_VAL(privkey)+1) == '/') {
+		newpath = zend_string_alloc(strlen(pws->pw_dir) + ZSTR_LEN(privkey), 0);
+		strcpy(ZSTR_VAL(newpath), pws->pw_dir);
+		strcat(ZSTR_VAL(newpath), ZSTR_VAL(privkey)+1);
+		zend_string_release(privkey);
 		privkey = newpath;
 	}
 #endif
 
 	/* TODO: Support passphrase callback */
-	if (libssh2_userauth_publickey_fromfile_ex(session, username, username_len, pubkey, privkey, passphrase)) {
+	if (libssh2_userauth_publickey_fromfile_ex(session, ZSTR_VAL(username), ZSTR_LEN(username), ZSTR_VAL(pubkey), ZSTR_VAL(privkey), ZSTR_VAL(passphrase))) {
 		char *buf;
 		int len;
 		libssh2_session_last_error(session, &buf, &len, 0);
-		php_error_docref(NULL, E_WARNING, "Authentication failed for %s using public key: %s", username, buf);
+		php_error_docref(NULL, E_WARNING, "Authentication failed for %s using public key: %s", ZSTR_VAL(username), buf);
 		RETURN_FALSE;
 	}
 
@@ -843,7 +843,7 @@ PHP_FUNCTION(ssh2_forward_accept)
  */
 PHP_FUNCTION(ssh2_poll)
 {
-	zval *zdesc, *subarray, ***pollmap;
+	zval *zdesc, *subarray, **pollmap;
 	zend_long timeout = PHP_SSH2_DEFAULT_POLL_TIMEOUT;
 	LIBSSH2_POLLFD *pollfds;
 	int numfds, i = 0, fds_ready;
@@ -883,7 +883,8 @@ PHP_FUNCTION(ssh2_poll)
 
 		pollfds[i].events = Z_LVAL_P(tmpzval);
 		hash_key_zstring = zend_string_init("resource", sizeof("resource") - 1, 0);
-		if ((tmpzval = zend_hash_find(Z_ARRVAL_P(subarray), hash_key_zstring)) == NULL || Z_TYPE_P(tmpzval) != IS_RESOURCE) {
+		if ((tmpzval = zend_hash_find(Z_ARRVAL_P(subarray), hash_key_zstring)) == NULL || Z_TYPE_P(tmpzval) != IS_REFERENCE
+			|| (tmpzval = Z_REFVAL_P(tmpzval)) == NULL || Z_TYPE_P(tmpzval) != IS_RESOURCE) {
 			php_error_docref(NULL, E_WARNING, "Invalid data in subarray, no resource element, or not of type resource");
 			numfds--;
 			zend_string_release(hash_key_zstring);
@@ -906,29 +907,23 @@ PHP_FUNCTION(ssh2_poll)
 			numfds--;
 			continue;
 		}
-		pollmap[i] = &subarray;
+		pollmap[i] = subarray;
 		i++;
 	}
 
 	fds_ready = libssh2_poll(pollfds, numfds, timeout * 1000);
 
 	for(i = 0; i < numfds; i++) {
-		zval *subarray = *pollmap[i];
+		zval *subarray = pollmap[i];
 
-		if (!Z_ISREF_P(subarray) && Z_REFCOUNT_P(subarray) > 1) {
+		if (Z_REFCOUNT_P(subarray) > 1) {
 			/* Make a new copy of the subarray zval* */
-			// TODO Sean-Der
-			//MAKE_STD_ZVAL(subarray);
-			*subarray = **pollmap[i];
+			zval new_subarray;
+			ZVAL_DUP(&new_subarray, subarray);
 
-			/* Point the pData to the new zval* and duplicate its resources */
-			*pollmap[i] = subarray;
-			zval_copy_ctor(subarray);
-
-			/* Fixup its refcount */
-			// TODO Sean-Der
-			//Z_UNSET_ISREF_P(subarray);
-			//Z_SET_REFCOUNT_P(subarray, 1);
+			/* Decrement refcount prior to swapping in new allocation */
+			Z_DELREF_P(subarray);
+			*pollmap[i] = new_subarray;
 		}
 		hash_key_zstring = zend_string_init("revents", sizeof("revents") - 1, 0);
 		zend_hash_del(Z_ARRVAL_P(subarray), hash_key_zstring);
@@ -1416,6 +1411,19 @@ ZEND_BEGIN_ARG_INFO_EX(arginfo_ssh2_auth_hostbased_file, 0, 0, 5)
  	ZEND_ARG_INFO(0, local_username)
 ZEND_END_ARG_INFO()
 
+ZEND_BEGIN_ARG_INFO_EX(arginfo_ssh2_forward_listen, 0, 0, 2)
+ 	ZEND_ARG_INFO(0, session)
+ 	ZEND_ARG_INFO(0, port)
+ 	ZEND_ARG_INFO(0, host)
+ 	ZEND_ARG_INFO(0, max_connections)
+ZEND_END_ARG_INFO()
+
+ZEND_BEGIN_ARG_INFO_EX(arginfo_ssh2_forward_accept, 0, 0, 1)
+ 	ZEND_ARG_INFO(0, listener)
+ 	ZEND_ARG_INFO(1, host)
+ 	ZEND_ARG_INFO(0, port)
+ZEND_END_ARG_INFO()
+
 ZEND_BEGIN_ARG_INFO_EX(arginfo_ssh2_shell, 0, 0, 1)
  	ZEND_ARG_INFO(0, session)
  	ZEND_ARG_INFO(0, termtype)
@@ -1459,6 +1467,15 @@ ZEND_BEGIN_ARG_INFO(arginfo_ssh2_fetch_stream, 2)
  	ZEND_ARG_INFO(0, streamid)
 ZEND_END_ARG_INFO()
 
+ZEND_BEGIN_ARG_INFO_EX(arginfo_ssh2_poll, 0, 0, 1)
+ 	ZEND_ARG_INFO(1, polldes)
+ 	ZEND_ARG_INFO(0, timeout)
+ZEND_END_ARG_INFO()
+
+ZEND_BEGIN_ARG_INFO_EX(arginfo_ssh2_send_eof, 0, 0, 1)
+	ZEND_ARG_INFO(0, channel)
+ZEND_END_ARG_INFO()
+
 ZEND_BEGIN_ARG_INFO(arginfo_ssh2_sftp, 1)
  	ZEND_ARG_INFO(0, session)
 ZEND_END_ARG_INFO()
@@ -1559,8 +1576,8 @@ zend_function_entry ssh2_functions[] = {
 	PHP_FE(ssh2_auth_pubkey_file,				arginfo_ssh2_auth_pubkey_file)
 	PHP_FE(ssh2_auth_hostbased_file,			arginfo_ssh2_auth_hostbased_file)
 
-	PHP_FE(ssh2_forward_listen,					NULL)
-	PHP_FE(ssh2_forward_accept,					NULL)
+	PHP_FE(ssh2_forward_listen,					arginfo_ssh2_forward_listen)
+	PHP_FE(ssh2_forward_accept,					arginfo_ssh2_forward_accept)
 
 	/* Stream Stuff */
 	PHP_FE(ssh2_shell,							arginfo_ssh2_shell)
@@ -1569,7 +1586,8 @@ zend_function_entry ssh2_functions[] = {
 	PHP_FE(ssh2_scp_recv,						arginfo_ssh2_scp_recv)
 	PHP_FE(ssh2_scp_send,						arginfo_ssh2_scp_send)
 	PHP_FE(ssh2_fetch_stream,					arginfo_ssh2_fetch_stream)
-	PHP_FE(ssh2_poll,							php_ssh2_first_arg_force_ref)
+	PHP_FE(ssh2_poll,							arginfo_ssh2_poll)
+	PHP_FE(ssh2_send_eof,						arginfo_ssh2_send_eof)
 
 	/* SFTP Stuff */
 	PHP_FE(ssh2_sftp,							arginfo_ssh2_sftp)
