@@ -40,10 +40,14 @@ void *php_ssh2_zval_from_resource_handle(int handle) {
    * channel_stream_ops *
    ********************** */
 
+#if PHP_VERSION_ID < 70400
 static size_t php_ssh2_channel_stream_write(php_stream *stream, const char *buf, size_t count)
+#else
+static ssize_t php_ssh2_channel_stream_write(php_stream *stream, const char *buf, size_t count)
+#endif
 {
 	php_ssh2_channel_data *abstract = (php_ssh2_channel_data*)stream->abstract;
-	size_t writestate;
+	ssize_t writestate;
 	LIBSSH2_SESSION *session;
 
 	libssh2_channel_set_blocking(abstract->channel, abstract->is_blocking);
@@ -64,24 +68,31 @@ static size_t php_ssh2_channel_stream_write(php_stream *stream, const char *buf,
 		libssh2_session_set_timeout(session, 0);
 	}
 #endif
-	if (writestate == LIBSSH2_ERROR_EAGAIN) {
-		writestate = 0;
-	}
 
-	if (writestate < 0) {
+	if (writestate == LIBSSH2_ERROR_EAGAIN) {
+#if PHP_VERSION_ID < 70400
+		writestate = 0;
+#endif
+	} else if (writestate < 0) {
 		char *error_msg = NULL;
 		if (libssh2_session_last_error(session, &error_msg, NULL, 0) == writestate) {
 			php_error_docref(NULL, E_WARNING, "Failure '%s' (%ld)", error_msg, writestate);
 		}
 
 		stream->eof = 1;
+#if PHP_VERSION_ID < 70400
 		writestate = 0;
+#endif
 	}
 
 	return writestate;
 }
 
+#if PHP_VERSION_ID < 70400
 static size_t php_ssh2_channel_stream_read(php_stream *stream, char *buf, size_t count)
+#else
+static ssize_t php_ssh2_channel_stream_read(php_stream *stream, char *buf, size_t count)
+#endif
 {
 	php_ssh2_channel_data *abstract = (php_ssh2_channel_data*)stream->abstract;
 	ssize_t readstate;
@@ -104,11 +115,12 @@ static size_t php_ssh2_channel_stream_read(php_stream *stream, char *buf, size_t
 		libssh2_session_set_timeout(session, 0);
 	}
 #endif
-	if (readstate == LIBSSH2_ERROR_EAGAIN) {
-		readstate = 0;
-	}
 
-	if (readstate < 0) {
+	if (readstate == LIBSSH2_ERROR_EAGAIN) {
+#if PHP_VERSION_ID < 70400
+		readstate = 0;
+#endif
+	} else if (readstate < 0) {
 		char *error_msg = NULL;
 		if (libssh2_session_last_error(session, &error_msg, NULL, 0) == readstate) {
 			php_error_docref(NULL, E_WARNING, "Failure '%s' (%ld)", error_msg, readstate);
@@ -145,6 +157,28 @@ static int php_ssh2_channel_stream_flush(php_stream *stream)
 	return libssh2_channel_flush_ex(abstract->channel, abstract->streamid);
 }
 
+static int php_ssh2_channel_stream_cast(php_stream *stream, int castas, void **ret)
+{
+	php_ssh2_channel_data *abstract = (php_ssh2_channel_data*)stream->abstract;
+	LIBSSH2_SESSION *session;
+	php_ssh2_session_data **session_data;
+
+	session = (LIBSSH2_SESSION *)zend_fetch_resource(abstract->session_rsrc, PHP_SSH2_SESSION_RES_NAME, le_ssh2_session);
+	session_data = (php_ssh2_session_data **)libssh2_session_abstract(session);
+
+	switch (castas)	{
+		case PHP_STREAM_AS_FD:
+		case PHP_STREAM_AS_FD_FOR_SELECT:
+		case PHP_STREAM_AS_SOCKETD:
+			if (ret) {
+				*(php_socket_t *)ret = (*session_data)->socket;
+			}
+			return SUCCESS;
+		default:
+			return FAILURE;
+	}
+}
+
 static int php_ssh2_channel_stream_set_option(php_stream *stream, int option, int value, void *ptrparam)
 {
 	php_ssh2_channel_data *abstract = (php_ssh2_channel_data*)stream->abstract;
@@ -187,7 +221,7 @@ php_stream_ops php_ssh2_channel_stream_ops = {
 	php_ssh2_channel_stream_flush,
 	PHP_SSH2_CHANNEL_STREAM_NAME,
 	NULL, /* seek */
-	NULL, /* cast */
+	php_ssh2_channel_stream_cast,
 	NULL, /* stat */
 	php_ssh2_channel_stream_set_option,
 };
@@ -270,10 +304,14 @@ php_url *php_ssh2_fopen_wraper_parse_path(const char *path, char *type, php_stre
 		php_ssh2_sftp_data *sftp_data;
 		zval *zresource;
 
-		zresource = php_ssh2_zval_from_resource_handle(resource_id);
+		if ((zresource = php_ssh2_zval_from_resource_handle(resource_id)) == NULL) {
+			php_url_free(resource);
+			return NULL;
+		}
 
 		if (psftp) {
-			sftp_data = (php_ssh2_sftp_data *)zend_fetch_resource(Z_RES_P(zresource), PHP_SSH2_SFTP_RES_NAME, le_ssh2_sftp);
+			/* suppress potential warning by passing NULL as resource_type_name */
+			sftp_data = (php_ssh2_sftp_data *)zend_fetch_resource(Z_RES_P(zresource), NULL, le_ssh2_sftp);
 			if (sftp_data) {
 				/* Want the sftp layer */
 				Z_ADDREF_P(zresource);
@@ -455,12 +493,13 @@ php_url *php_ssh2_fopen_wraper_parse_path(const char *path, char *type, php_stre
 
 	/* Auth failure */
 	php_url_free(resource);
-	zend_list_delete(Z_RES(zsession));
+	if (Z_RES(zsession)) {
+		zend_list_delete(Z_RES(zsession));
+	}
 	return NULL;
 
 session_authed:
-	//TODO Sean-Der
-	//ZEND_REGISTER_RESOURCE(&zsession, session, le_ssh2_session);
+	ZVAL_RES(&zsession, zend_register_resource(session, le_ssh2_session));
 
 	if (psftp) {
 		LIBSSH2_SFTP *sftp;
@@ -1405,6 +1444,42 @@ PHP_FUNCTION(ssh2_fetch_stream)
 }
 /* }}} */
 
+/* {{{ proto stream ssh2_send_eof(stream channel)
+ * Sends EOF to a stream. Primary use is to close stdin of an stdio stream.
+ */
+PHP_FUNCTION(ssh2_send_eof)
+{
+	php_ssh2_channel_data *data;
+	php_stream *parent;
+	zval *zparent;
+	int ssh2_ret;
+
+	if (zend_parse_parameters(ZEND_NUM_ARGS(), "r", &zparent) == FAILURE) {
+		return;
+	}
+
+	php_stream_from_zval(parent, zparent);
+	if (parent->ops != &php_ssh2_channel_stream_ops) {
+		php_error_docref(NULL, E_WARNING, "Provided stream is not of type " PHP_SSH2_CHANNEL_STREAM_NAME);
+		RETURN_FALSE;
+	}
+
+	data = (php_ssh2_channel_data*)parent->abstract;
+	if (!data) {
+		php_error_docref(NULL, E_WARNING, "Abstract in stream is null");
+		RETURN_FALSE;
+	}
+
+	ssh2_ret = libssh2_channel_send_eof(data->channel);
+	if (ssh2_ret < 0) {
+		php_error_docref(NULL, E_WARNING, "Couldn't send EOF to channel (Return code %d)", ssh2_ret);
+		RETURN_FALSE;
+	}
+
+	RETURN_TRUE;
+}
+/* }}} */
+
 /*
  * Local variables:
  * tab-width: 4
