@@ -1,19 +1,21 @@
 /*
-   +----------------------------------------------------------------------+
-   | PHP Version 7                                                        |
-   +----------------------------------------------------------------------+
-   | Copyright (c) 1997-2016 The PHP Group                                |
-   +----------------------------------------------------------------------+
-   | This source file is subject to version 3.01 of the PHP license,      |
-   | that is bundled with this package in the file LICENSE, and is        |
-   | available through the world-wide-web at the following url:           |
-   | http://www.php.net/license/3_01.txt                                  |
-   | If you did not receive a copy of the PHP license and are unable to   |
-   | obtain it through the world-wide-web, please send a note to          |
-   | license@php.net so we can mail you a copy immediately.               |
-   +----------------------------------------------------------------------+
-   | Author: Sara Golemon <pollita@php.net>                               |
-   +----------------------------------------------------------------------+
+  +----------------------------------------------------------------------+
+  | PHP Version 4                                                        |
+  +----------------------------------------------------------------------+
+  | Copyright (c) 1997-2006 The PHP Group                                |
+  +----------------------------------------------------------------------+
+  | This source file is subject to version 3.01 of the PHP license,      |
+  | that is bundled with this package in the file LICENSE, and is        |
+  | available at through the world-wide-web at                           |
+  | http://www.php.net/license/3_01.txt.                                 |
+  | If you did not receive a copy of the PHP license and are unable to   |
+  | obtain it through the world-wide-web, please send a note to          |
+  | license@php.net so we can mail you a copy immediately.               |
+  +----------------------------------------------------------------------+
+  | Author: Sara Golemon <pollita@php.net>                               |
+  +----------------------------------------------------------------------+
+
+  $Id$
 */
 
 #ifdef HAVE_CONFIG_H
@@ -28,7 +30,7 @@
    * Resource Housekeeping *
    ************************* */
 
-void php_ssh2_sftp_dtor(zend_resource *rsrc)
+void php_ssh2_sftp_dtor(zend_rsrc_list_entry *rsrc TSRMLS_DC)
 {
 	php_ssh2_sftp_data *data = (php_ssh2_sftp_data*)rsrc->ptr;
 
@@ -38,7 +40,7 @@ void php_ssh2_sftp_dtor(zend_resource *rsrc)
 
 	libssh2_sftp_shutdown(data->sftp);
 
-	zend_list_delete(data->session_rsrc);
+	zend_list_delete(data->session_rsrcid);
 
 	efree(data);
 }
@@ -47,7 +49,7 @@ void php_ssh2_sftp_dtor(zend_resource *rsrc)
    * SFTP File Ops *
    ***************** */
 
-unsigned long php_ssh2_parse_fopen_modes(char *openmode) {
+inline unsigned long php_ssh2_parse_fopen_modes(char *openmode) {
 	unsigned long flags = 0;
 
 	if (strchr(openmode, 'a')) {
@@ -73,13 +75,13 @@ unsigned long php_ssh2_parse_fopen_modes(char *openmode) {
 	return flags;
 }
 
-static inline int php_ssh2_sftp_attr2ssb(php_stream_statbuf *ssb, LIBSSH2_SFTP_ATTRIBUTES *attrs)
+inline int php_ssh2_sftp_attr2ssb(php_stream_statbuf *ssb, LIBSSH2_SFTP_ATTRIBUTES *attrs)
 {
 	memset(ssb, 0, sizeof(php_stream_statbuf));
 	if (attrs->flags & LIBSSH2_SFTP_ATTR_SIZE) {
 		ssb->sb.st_size = attrs->filesize;
 	}
-
+	
 	if (attrs->flags & LIBSSH2_SFTP_ATTR_UIDGID) {
 		ssb->sb.st_uid = attrs->uid;
 		ssb->sb.st_gid = attrs->gid;
@@ -98,12 +100,12 @@ static inline int php_ssh2_sftp_attr2ssb(php_stream_statbuf *ssb, LIBSSH2_SFTP_A
 typedef struct _php_ssh2_sftp_handle_data {
 	LIBSSH2_SFTP_HANDLE *handle;
 
-	zend_resource *sftp_rsrc;
+	long sftp_rsrcid;
 } php_ssh2_sftp_handle_data;
 
 /* {{{ php_ssh2_sftp_stream_write
  */
-static size_t php_ssh2_sftp_stream_write(php_stream *stream, const char *buf, size_t count)
+static size_t php_ssh2_sftp_stream_write(php_stream *stream, const char *buf, size_t count TSRMLS_DC)
 {
 	php_ssh2_sftp_handle_data *data = (php_ssh2_sftp_handle_data*)stream->abstract;
 	ssize_t bytes_written;
@@ -116,7 +118,7 @@ static size_t php_ssh2_sftp_stream_write(php_stream *stream, const char *buf, si
 
 /* {{{ php_ssh2_sftp_stream_read
  */
-static size_t php_ssh2_sftp_stream_read(php_stream *stream, char *buf, size_t count)
+static size_t php_ssh2_sftp_stream_read(php_stream *stream, char *buf, size_t count TSRMLS_DC)
 {
 	php_ssh2_sftp_handle_data *data = (php_ssh2_sftp_handle_data*)stream->abstract;
 	ssize_t bytes_read;
@@ -131,12 +133,12 @@ static size_t php_ssh2_sftp_stream_read(php_stream *stream, char *buf, size_t co
 
 /* {{{ php_ssh2_sftp_stream_close
  */
-static int php_ssh2_sftp_stream_close(php_stream *stream, int close_handle)
+static int php_ssh2_sftp_stream_close(php_stream *stream, int close_handle TSRMLS_DC)
 {
 	php_ssh2_sftp_handle_data *data = (php_ssh2_sftp_handle_data*)stream->abstract;
 
 	libssh2_sftp_close(data->handle);
-	zend_list_delete(data->sftp_rsrc);
+	zend_list_delete(data->sftp_rsrcid);
 	efree(data);
 
 	return 0;
@@ -145,7 +147,7 @@ static int php_ssh2_sftp_stream_close(php_stream *stream, int close_handle)
 
 /* {{{ php_ssh2_sftp_stream_seek
  */
-static int php_ssh2_sftp_stream_seek(php_stream *stream, zend_off_t offset, int whence, zend_off_t *newoffset)
+static int php_ssh2_sftp_stream_seek(php_stream *stream, off_t offset, int whence, off_t *newoffset TSRMLS_DC)
 {
 	php_ssh2_sftp_handle_data *data = (php_ssh2_sftp_handle_data*)stream->abstract;
 
@@ -164,7 +166,7 @@ static int php_ssh2_sftp_stream_seek(php_stream *stream, zend_off_t offset, int
 		}
 		case SEEK_CUR:
 		{
-			zend_off_t current_offset = libssh2_sftp_tell(data->handle);
+			off_t current_offset = libssh2_sftp_tell(data->handle);
 
 			if (current_offset < 0) {
 				return -1;
@@ -186,7 +188,7 @@ static int php_ssh2_sftp_stream_seek(php_stream *stream, zend_off_t offset, int
 
 /* {{{ php_ssh2_sftp_stream_fstat
  */
-static int php_ssh2_sftp_stream_fstat(php_stream *stream, php_stream_statbuf *ssb)
+static int php_ssh2_sftp_stream_fstat(php_stream *stream, php_stream_statbuf *ssb TSRMLS_DC)
 {
 	php_ssh2_sftp_handle_data *data = (php_ssh2_sftp_handle_data*)stream->abstract;
 	LIBSSH2_SFTP_ATTRIBUTES attrs;
@@ -200,56 +202,55 @@ static int php_ssh2_sftp_stream_fstat(php_stream *stream, php_stream_statbuf *ss
 /* }}} */
 
 static php_stream_ops php_ssh2_sftp_stream_ops = {
-	php_ssh2_sftp_stream_write,
-	php_ssh2_sftp_stream_read,
-	php_ssh2_sftp_stream_close,
-	NULL, /* flush */
-	PHP_SSH2_SFTP_STREAM_NAME,
-	php_ssh2_sftp_stream_seek,
-	NULL, /* cast */
-	php_ssh2_sftp_stream_fstat,
+    php_ssh2_sftp_stream_write,
+    php_ssh2_sftp_stream_read,
+    php_ssh2_sftp_stream_close,
+    NULL, /* flush */
+    PHP_SSH2_SFTP_STREAM_NAME,
+    php_ssh2_sftp_stream_seek,
+    NULL, /* cast */
+    php_ssh2_sftp_stream_fstat,
 	NULL, /* set_option */
 };
 
 /* {{{ php_ssh2_sftp_stream_opener
  */
-
-static php_stream *php_ssh2_sftp_stream_opener(php_stream_wrapper *wrapper, const char *filename, const char *mode,
-		int options, zend_string **opened_path, php_stream_context *context STREAMS_DC)
+static php_stream *php_ssh2_sftp_stream_opener(	php_stream_wrapper *wrapper, char *filename, char *mode,
+												int options, char **opened_path, php_stream_context *context STREAMS_DC TSRMLS_DC)
 {
 	php_ssh2_sftp_handle_data *data;
 	LIBSSH2_SESSION *session = NULL;
 	LIBSSH2_SFTP *sftp = NULL;
 	LIBSSH2_SFTP_HANDLE *handle;
 	php_stream *stream;
-	zend_resource *rsrc = NULL, *sftp_rsrc = NULL;
+	int resource_id = 0, sftp_rsrcid = 0;
 	php_url *resource;
 	unsigned long flags;
 	long perms = 0644;
 
-	resource = php_ssh2_fopen_wraper_parse_path(filename, "sftp", context, &session, &rsrc, &sftp, &sftp_rsrc);
+	resource = php_ssh2_fopen_wraper_parse_path(filename, "sftp", context, &session, &resource_id, &sftp, &sftp_rsrcid TSRMLS_CC);
 	if (!resource || !session || !sftp) {
 		return NULL;
 	}
 
-	flags = php_ssh2_parse_fopen_modes((char *)mode);
+	flags = php_ssh2_parse_fopen_modes(mode);
 
-	handle = libssh2_sftp_open(sftp, SSH2_URL_STR(resource->path), flags, perms);
+	handle = libssh2_sftp_open(sftp, resource->path, flags, perms);
 	if (!handle) {
-		php_error_docref(NULL, E_WARNING, "Unable to open %s on remote host", filename);
+		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Unable to open %s on remote host", filename);
 		php_url_free(resource);
-		zend_list_delete(sftp_rsrc);
+		zend_list_delete(sftp_rsrcid);
 		return NULL;
 	}
 
 	data = emalloc(sizeof(php_ssh2_sftp_handle_data));
 	data->handle = handle;
-	data->sftp_rsrc = sftp_rsrc;
+	data->sftp_rsrcid = sftp_rsrcid;
 
 	stream = php_stream_alloc(&php_ssh2_sftp_stream_ops, data, 0, mode);
 	if (!stream) {
 		libssh2_sftp_close(handle);
-		zend_list_delete(sftp_rsrc);
+		zend_list_delete(sftp_rsrcid);
 		efree(data);
 	}
 	php_url_free(resource);
@@ -264,27 +265,40 @@ static php_stream *php_ssh2_sftp_stream_opener(php_stream_wrapper *wrapper, cons
 
 /* {{{ php_ssh2_sftp_dirstream_read
  */
-static size_t php_ssh2_sftp_dirstream_read(php_stream *stream, char *buf, size_t count)
+static size_t php_ssh2_sftp_dirstream_read(php_stream *stream, char *buf, size_t count TSRMLS_DC)
 {
 	php_ssh2_sftp_handle_data *data = (php_ssh2_sftp_handle_data*)stream->abstract;
 	php_stream_dirent *ent = (php_stream_dirent*)buf;
 	int bytesread = libssh2_sftp_readdir(data->handle, ent->d_name, sizeof(ent->d_name) - 1, NULL);
-	zend_string *basename;
+	char *basename = NULL;
+	size_t basename_len = 0;
 
 	if (bytesread <= 0) {
 		return 0;
 	}
 	ent->d_name[bytesread] = 0;
 
+#ifdef ZEND_ENGINE_2
+	php_basename(ent->d_name, bytesread, NULL, 0, &basename, &basename_len TSRMLS_CC);
+#else
+/* CURSE YOU BC BREAKS! */
 	basename = php_basename(ent->d_name, bytesread, NULL, 0);
+	if (basename) {
+		basename_len = strlen(basename);
+	}
+#endif
 	if (!basename) {
 		return 0;
 	}
 
-	bytesread = MIN(sizeof(ent->d_name) - 1, basename->len);
-	memcpy(ent->d_name, basename->val, bytesread);
+	if (!basename_len) {
+		efree(basename);
+		return 0;
+	}
+	bytesread = MIN(sizeof(ent->d_name) - 1, basename_len);
+	memcpy(ent->d_name, basename, bytesread);
 	ent->d_name[bytesread] = 0;
-	zend_string_release(basename);
+	efree(basename);
 
 	return sizeof(php_stream_dirent);
 }
@@ -292,12 +306,12 @@ static size_t php_ssh2_sftp_dirstream_read(php_stream *stream, char *buf, size_t
 
 /* {{{ php_ssh2_sftp_dirstream_close
  */
-static int php_ssh2_sftp_dirstream_close(php_stream *stream, int close_handle)
+static int php_ssh2_sftp_dirstream_close(php_stream *stream, int close_handle TSRMLS_DC)
 {
 	php_ssh2_sftp_handle_data *data = (php_ssh2_sftp_handle_data*)stream->abstract;
 
 	libssh2_sftp_close(data->handle);
-	zend_list_delete(data->sftp_rsrc);
+	zend_list_delete(data->sftp_rsrcid);
 	efree(data);
 
 	return 0;
@@ -306,50 +320,50 @@ static int php_ssh2_sftp_dirstream_close(php_stream *stream, int close_handle)
 
 static php_stream_ops php_ssh2_sftp_dirstream_ops = {
 	NULL, /* write */
-	php_ssh2_sftp_dirstream_read,
-	php_ssh2_sftp_dirstream_close,
-	NULL, /* flush */
-	PHP_SSH2_SFTP_DIRSTREAM_NAME,
+    php_ssh2_sftp_dirstream_read,
+    php_ssh2_sftp_dirstream_close,
+    NULL, /* flush */
+    PHP_SSH2_SFTP_DIRSTREAM_NAME,
 	NULL, /* seek */
-	NULL, /* cast */
+    NULL, /* cast */
 	NULL, /* fstat */
 	NULL, /* set_option */
 };
 
 /* {{{ php_ssh2_sftp_dirstream_opener
  */
-static php_stream *php_ssh2_sftp_dirstream_opener(php_stream_wrapper *wrapper, const char *filename, const char *mode,
-		int options, zend_string **opened_path, php_stream_context *context STREAMS_DC)
+static php_stream *php_ssh2_sftp_dirstream_opener(	php_stream_wrapper *wrapper, char *filename, char *mode,
+													int options, char **opened_path, php_stream_context *context STREAMS_DC TSRMLS_DC)
 {
 	php_ssh2_sftp_handle_data *data;
 	LIBSSH2_SESSION *session = NULL;
 	LIBSSH2_SFTP *sftp = NULL;
 	LIBSSH2_SFTP_HANDLE *handle;
 	php_stream *stream;
-	zend_resource *rsrc = NULL, *sftp_rsrc = NULL;
+	int resource_id = 0, sftp_rsrcid = 0;
 	php_url *resource;
 
-	resource = php_ssh2_fopen_wraper_parse_path(filename, "sftp", context, &session, &rsrc, &sftp, &sftp_rsrc);
+	resource = php_ssh2_fopen_wraper_parse_path(filename, "sftp", context, &session, &resource_id, &sftp, &sftp_rsrcid TSRMLS_CC);
 	if (!resource || !session || !sftp) {
 		return NULL;
 	}
 
-	handle = libssh2_sftp_opendir(sftp, SSH2_URL_STR(resource->path));
+	handle = libssh2_sftp_opendir(sftp, resource->path);
 	if (!handle) {
-		php_error_docref(NULL, E_WARNING, "Unable to open %s on remote host", filename);
+		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Unable to open %s on remote host", filename);
 		php_url_free(resource);
-		zend_list_delete(sftp_rsrc);
+		zend_list_delete(sftp_rsrcid);
 		return NULL;
 	}
 
 	data = emalloc(sizeof(php_ssh2_sftp_handle_data));
 	data->handle = handle;
-	data->sftp_rsrc = sftp_rsrc;
+	data->sftp_rsrcid = sftp_rsrcid;
 
 	stream = php_stream_alloc(&php_ssh2_sftp_dirstream_ops, data, 0, mode);
 	if (!stream) {
 		libssh2_sftp_close(handle);
-		zend_list_delete(sftp_rsrc);
+		zend_list_delete(sftp_rsrcid);
 		efree(data);
 	}
 	php_url_free(resource);
@@ -362,32 +376,33 @@ static php_stream *php_ssh2_sftp_dirstream_opener(php_stream_wrapper *wrapper, c
    * SFTP Wrapper *
    **************** */
 
+#ifdef ZEND_ENGINE_2
 /* {{{ php_ssh2_sftp_urlstat
  */
-static int php_ssh2_sftp_urlstat(php_stream_wrapper *wrapper, const char *url, int flags, php_stream_statbuf *ssb, php_stream_context *context)
+static int php_ssh2_sftp_urlstat(php_stream_wrapper *wrapper, char *url, int flags, php_stream_statbuf *ssb, php_stream_context *context TSRMLS_DC)
 {
 	LIBSSH2_SFTP_ATTRIBUTES attrs;
 	LIBSSH2_SESSION *session = NULL;
 	LIBSSH2_SFTP *sftp = NULL;
-	zend_resource *rsrc = NULL, *sftp_rsrc = NULL;
+	int resource_id = 0, sftp_rsrcid = 0;
 	php_url *resource;
 
-	resource = php_ssh2_fopen_wraper_parse_path(url, "sftp", context, &session, &rsrc, &sftp, &sftp_rsrc);
+	resource = php_ssh2_fopen_wraper_parse_path(url, "sftp", context, &session, &resource_id, &sftp, &sftp_rsrcid TSRMLS_CC);
 	if (!resource || !session || !sftp || !resource->path) {
 		return -1;
 	}
 
-	if (libssh2_sftp_stat_ex(sftp, SSH2_URL_STR(resource->path), SSH2_URL_LEN(resource->path),
+	if (libssh2_sftp_stat_ex(sftp, resource->path, strlen(resource->path),
 		(flags & PHP_STREAM_URL_STAT_LINK) ? LIBSSH2_SFTP_LSTAT : LIBSSH2_SFTP_STAT, &attrs)) {
 		php_url_free(resource);
-		//zend_list_delete(sftp_rsrcid);
+		zend_list_delete(sftp_rsrcid);
 		return -1;
 	}
-
+	
 	php_url_free(resource);
 
 	/* parse_path addrefs the resource, but we're not holding on to it so we have to delref it before we leave */
-	//zend_list_delete(sftp_rsrcid);
+	zend_list_delete(sftp_rsrcid);
 
 	return php_ssh2_sftp_attr2ssb(ssb, &attrs);
 }
@@ -395,15 +410,15 @@ static int php_ssh2_sftp_urlstat(php_stream_wrapper *wrapper, const char *url, i
 
 /* {{{ php_ssh2_sftp_unlink
  */
-static int php_ssh2_sftp_unlink(php_stream_wrapper *wrapper, const char *url, int options, php_stream_context *context)
+static int php_ssh2_sftp_unlink(php_stream_wrapper *wrapper, char *url, int options, php_stream_context *context TSRMLS_DC)
 {
 	LIBSSH2_SESSION *session = NULL;
 	LIBSSH2_SFTP *sftp = NULL;
-	zend_resource *rsrc = NULL, *sftp_rsrc = NULL;
+	int resource_id = 0, sftp_rsrcid = 0;
 	php_url *resource;
 	int result;
 
-	resource = php_ssh2_fopen_wraper_parse_path(url, "sftp", context, &session, &rsrc, &sftp, &sftp_rsrc);
+	resource = php_ssh2_fopen_wraper_parse_path(url, "sftp", context, &session, &resource_id, &sftp, &sftp_rsrcid TSRMLS_CC);
 	if (!resource || !session || !sftp || !resource->path) {
 		if (resource) {
 			php_url_free(resource);
@@ -411,10 +426,10 @@ static int php_ssh2_sftp_unlink(php_stream_wrapper *wrapper, const char *url, in
 		return 0;
 	}
 
-	result = libssh2_sftp_unlink(sftp, SSH2_URL_STR(resource->path));
+	result = libssh2_sftp_unlink(sftp, resource->path);
 	php_url_free(resource);
 
-	//zend_list_delete(sftp_rsrcid);
+	zend_list_delete(sftp_rsrcid);
 
 	/* libssh2 uses 0 for success and the streams API uses 0 for failure, so invert */
 	return (result == 0) ? -1 : 0;
@@ -423,11 +438,11 @@ static int php_ssh2_sftp_unlink(php_stream_wrapper *wrapper, const char *url, in
 
 /* {{{ php_ssh2_sftp_rename
  */
-static int php_ssh2_sftp_rename(php_stream_wrapper *wrapper, const char *url_from, const char *url_to, int options, php_stream_context *context)
+static int php_ssh2_sftp_rename(php_stream_wrapper *wrapper, char *url_from, char *url_to, int options, php_stream_context *context TSRMLS_DC)
 {
 	LIBSSH2_SESSION *session = NULL;
 	LIBSSH2_SFTP *sftp = NULL;
-	zend_resource *rsrc = NULL, *sftp_rsrc = NULL;
+	int resource_id = 0, sftp_rsrcid = 0;
 	php_url *resource, *resource_to;
 	int result;
 
@@ -444,7 +459,7 @@ static int php_ssh2_sftp_rename(php_stream_wrapper *wrapper, const char *url_fro
 		return 0;
 	}
 
-	resource = php_ssh2_fopen_wraper_parse_path(url_from, "sftp", context, &session, &rsrc, &sftp, &sftp_rsrc);
+	resource = php_ssh2_fopen_wraper_parse_path(url_from, "sftp", context, &session, &resource_id, &sftp, &sftp_rsrcid TSRMLS_CC);
 	if (!resource || !session || !sftp || !resource->path) {
 		if (resource) {
 			php_url_free(resource);
@@ -453,11 +468,11 @@ static int php_ssh2_sftp_rename(php_stream_wrapper *wrapper, const char *url_fro
 		return 0;
 	}
 
-	result = libssh2_sftp_rename(sftp, SSH2_URL_STR(resource->path), SSH2_URL_STR(resource_to->path));
+	result = libssh2_sftp_rename(sftp, resource->path, resource_to->path);
 	php_url_free(resource);
 	php_url_free(resource_to);
 
-	//zend_list_delete(sftp_rsrcid);
+	zend_list_delete(sftp_rsrcid);
 
 	/* libssh2 uses 0 for success and the streams API uses 0 for failure, so invert */
 	return (result == 0) ? -1 : 0;
@@ -466,15 +481,15 @@ static int php_ssh2_sftp_rename(php_stream_wrapper *wrapper, const char *url_fro
 
 /* {{{ php_ssh2_sftp_mkdir
  */
-static int php_ssh2_sftp_mkdir(php_stream_wrapper *wrapper, const char *url, int mode, int options, php_stream_context *context)
+static int php_ssh2_sftp_mkdir(php_stream_wrapper *wrapper, char *url, int mode, int options, php_stream_context *context TSRMLS_DC)
 {
 	LIBSSH2_SESSION *session = NULL;
 	LIBSSH2_SFTP *sftp = NULL;
-	zend_resource *rsrc = NULL, *sftp_rsrc = NULL;
+	int resource_id = 0, sftp_rsrcid = 0;
 	php_url *resource;
 	int result;
 
-	resource = php_ssh2_fopen_wraper_parse_path(url, "sftp", context, &session, &rsrc, &sftp, &sftp_rsrc);
+	resource = php_ssh2_fopen_wraper_parse_path(url, "sftp", context, &session, &resource_id, &sftp, &sftp_rsrcid TSRMLS_CC);
 	if (!resource || !session || !sftp || !resource->path) {
 		if (resource) {
 			php_url_free(resource);
@@ -484,16 +499,16 @@ static int php_ssh2_sftp_mkdir(php_stream_wrapper *wrapper, const char *url, int
 
 	if (options & PHP_STREAM_MKDIR_RECURSIVE) {
 		/* Just attempt to make every directory, some will fail, but we only care about the last success/failure */
-		char *p = SSH2_URL_STR(resource->path);
+		char *p = resource->path;
 		while ((p = strchr(p + 1, '/'))) {
-			libssh2_sftp_mkdir_ex(sftp, SSH2_URL_STR(resource->path), p - SSH2_URL_STR(resource->path), mode);
+			libssh2_sftp_mkdir_ex(sftp, resource->path, p - resource->path, mode);
 		}
 	}
 
-	result = libssh2_sftp_mkdir(sftp, SSH2_URL_STR(resource->path), mode);
+	result = libssh2_sftp_mkdir(sftp, resource->path, mode);
 	php_url_free(resource);
 
-	//zend_list_delete(sftp_rsrcid);
+	zend_list_delete(sftp_rsrcid);
 
 	/* libssh2 uses 0 for success and the streams API uses 0 for failure, so invert */
 	return (result == 0) ? -1 : 0;
@@ -502,15 +517,15 @@ static int php_ssh2_sftp_mkdir(php_stream_wrapper *wrapper, const char *url, int
 
 /* {{{ php_ssh2_sftp_rmdir
  */
-static int php_ssh2_sftp_rmdir(php_stream_wrapper *wrapper, const char *url, int options, php_stream_context *context)
+static int php_ssh2_sftp_rmdir(php_stream_wrapper *wrapper, char *url, int options, php_stream_context *context TSRMLS_DC)
 {
 	LIBSSH2_SESSION *session = NULL;
 	LIBSSH2_SFTP *sftp = NULL;
-	zend_resource *rsrc = NULL, *sftp_rsrc = NULL;
+	int resource_id = 0, sftp_rsrcid = 0;
 	php_url *resource;
 	int result;
 
-	resource = php_ssh2_fopen_wraper_parse_path(url, "sftp", context, &session, &rsrc, &sftp, &sftp_rsrc);
+	resource = php_ssh2_fopen_wraper_parse_path(url, "sftp", context, &session, &resource_id, &sftp, &sftp_rsrcid TSRMLS_CC);
 	if (!resource || !session || !sftp || !resource->path) {
 		if (resource) {
 			php_url_free(resource);
@@ -518,33 +533,44 @@ static int php_ssh2_sftp_rmdir(php_stream_wrapper *wrapper, const char *url, int
 		return 0;
 	}
 
-	result = libssh2_sftp_rmdir(sftp, SSH2_URL_STR(resource->path));
+	result = libssh2_sftp_rmdir(sftp, resource->path);
 	php_url_free(resource);
 
-	//zend_list_delete(sftp_rsrcid);
+	zend_list_delete(sftp_rsrcid);
 
 	/* libssh2 uses 0 for success and the streams API uses 0 for failure, so invert */
 	return (result == 0) ? -1 : 0;
 }
 /* }}} */
+#endif
 
 static php_stream_wrapper_ops php_ssh2_sftp_wrapper_ops = {
 	php_ssh2_sftp_stream_opener,
 	NULL, /* close */
 	NULL, /* stat */
+#ifdef ZEND_ENGINE_2
 	php_ssh2_sftp_urlstat,
+#else
+	NULL, /* url_stat isn't actually functional prior to PHP5 */
+#endif
 	php_ssh2_sftp_dirstream_opener,
 	PHP_SSH2_SFTP_WRAPPER_NAME,
+#ifdef ZEND_ENGINE_2
 	php_ssh2_sftp_unlink,
 	php_ssh2_sftp_rename,
 	php_ssh2_sftp_mkdir,
 	php_ssh2_sftp_rmdir,
+#endif
 };
 
 php_stream_wrapper php_ssh2_sftp_wrapper = {
 	&php_ssh2_sftp_wrapper_ops,
 	NULL,
 	1,
+#if PHP_MAJOR_VERSION <= 5 && PHP_MINOR_VERSION <= 4
+	0,
+	NULL,
+#endif
 };
 
 /* *****************
@@ -562,30 +588,28 @@ PHP_FUNCTION(ssh2_sftp)
 	php_ssh2_sftp_data *data;
 	zval *zsession;
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), "r", &zsession) == FAILURE) {
+	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "r", &zsession) == FAILURE) {
 		return;
 	}
 
-	if ((session = (LIBSSH2_SESSION *)zend_fetch_resource(Z_RES_P(zsession), PHP_SSH2_SESSION_RES_NAME, le_ssh2_session)) == NULL) {
-		RETURN_FALSE;
-	}
+	ZEND_FETCH_RESOURCE(session, LIBSSH2_SESSION*, &zsession, -1, PHP_SSH2_SESSION_RES_NAME, le_ssh2_session);
 
 	sftp = libssh2_sftp_init(session);
 	if (!sftp) {
 		char *sess_err = "Unknown";
 
 		libssh2_session_last_error(session, &sess_err, NULL, 0);
-		php_error_docref(NULL, E_WARNING, "Unable to startup SFTP subsystem: %s", sess_err);
+		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Unable to startup SFTP subsystem: %s", sess_err);
 		RETURN_FALSE;
 	}
 
 	data = emalloc(sizeof(php_ssh2_sftp_data));
 	data->session = session;
 	data->sftp = sftp;
-	data->session_rsrc = Z_RES_P(zsession);
-	Z_ADDREF_P(zsession);
+	data->session_rsrcid = Z_LVAL_P(zsession);
+	zend_list_addref(Z_LVAL_P(zsession));
 
-	RETURN_RES(zend_register_resource(data, le_ssh2_sftp));
+	ZEND_REGISTER_RESOURCE(return_value, data, le_ssh2_sftp);
 }
 /* }}} */
 
@@ -597,17 +621,16 @@ PHP_FUNCTION(ssh2_sftp_rename)
 {
 	php_ssh2_sftp_data *data;
 	zval *zsftp;
-	zend_string *src, *dst;
+	char *src, *dst;
+	int src_len, dst_len;
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), "rSS", &zsftp, &src, &dst) == FAILURE) {
+	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rss", &zsftp, &src, &src_len, &dst, &dst_len) == FAILURE) {
 		return;
 	}
 
-	if ((data = (php_ssh2_sftp_data *)zend_fetch_resource(Z_RES_P(zsftp), PHP_SSH2_SFTP_RES_NAME, le_ssh2_sftp)) == NULL) {
-		RETURN_FALSE;
-	}
+	ZEND_FETCH_RESOURCE(data, php_ssh2_sftp_data*, &zsftp, -1, PHP_SSH2_SFTP_RES_NAME, le_ssh2_sftp);
 
-	RETURN_BOOL(!libssh2_sftp_rename_ex(data->sftp, src->val, src->len, dst->val, dst->len,
+	RETURN_BOOL(!libssh2_sftp_rename_ex(data->sftp, src, src_len, dst, dst_len,
 				 LIBSSH2_SFTP_RENAME_OVERWRITE | LIBSSH2_SFTP_RENAME_ATOMIC | LIBSSH2_SFTP_RENAME_NATIVE));
 }
 /* }}} */
@@ -618,76 +641,73 @@ PHP_FUNCTION(ssh2_sftp_unlink)
 {
 	php_ssh2_sftp_data *data;
 	zval *zsftp;
-	zend_string *filename;
+	char *filename;
+	int filename_len;
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), "rS", &zsftp, &filename) == FAILURE) {
+	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rs", &zsftp, &filename, &filename_len) == FAILURE) {
 		return;
 	}
 
-	if ((data = (php_ssh2_sftp_data *)zend_fetch_resource(Z_RES_P(zsftp), PHP_SSH2_SFTP_RES_NAME, le_ssh2_sftp)) == NULL) {
-		RETURN_FALSE;
-	}
+	ZEND_FETCH_RESOURCE(data, php_ssh2_sftp_data*, &zsftp, -1, PHP_SSH2_SFTP_RES_NAME, le_ssh2_sftp);
 
-	RETURN_BOOL(!libssh2_sftp_unlink_ex(data->sftp, filename->val, filename->len));
+	RETURN_BOOL(!libssh2_sftp_unlink_ex(data->sftp, filename, filename_len));
 }
 /* }}} */
 
-/* {{{ proto bool ssh2_sftp_mkdir(resource sftp, string dirname[, int mode[, int recursive]])
+/* {{{ proto bool ssh2_sftp_mkdir(resource sftp, string filename[, int mode[, int recursive]])
  */
 PHP_FUNCTION(ssh2_sftp_mkdir)
 {
 	php_ssh2_sftp_data *data;
 	zval *zsftp;
-	zend_string *dirname;
-	zend_long mode = 0777;
+	char *filename;
+	int filename_len;
+	long mode = 0777;
 	zend_bool recursive = 0;
 	char *p;
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), "rS|lb", &zsftp, &dirname, &mode, &recursive) == FAILURE) {
+	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rs|lb", &zsftp, &filename, &filename_len, &mode, &recursive) == FAILURE) {
 		return;
 	}
 
-	if (!dirname) {
+	if (filename_len < 1) {
 		RETURN_FALSE;
 	}
 
-	if ((data = (php_ssh2_sftp_data *)zend_fetch_resource(Z_RES_P(zsftp), PHP_SSH2_SFTP_RES_NAME, le_ssh2_sftp)) == NULL) {
-		RETURN_FALSE;
-	}
+	ZEND_FETCH_RESOURCE(data, php_ssh2_sftp_data*, &zsftp, -1, PHP_SSH2_SFTP_RES_NAME, le_ssh2_sftp);
 
 	if (recursive) {
 		/* Just attempt to make every directory, some will fail, but we only care about the last success/failure */
-		p = dirname->val;
+		p = filename;
 		while ((p = strchr(p + 1, '/'))) {
-			if ((p - dirname->val) + 1 == dirname->len) {
+			if ((p - filename) + 1 == filename_len) {
 				break;
 			}
-			libssh2_sftp_mkdir_ex(data->sftp, dirname->val, p - dirname->val, mode);
+			libssh2_sftp_mkdir_ex(data->sftp, filename, p - filename, mode);
 		}
 	}
 
 
-	RETURN_BOOL(!libssh2_sftp_mkdir_ex(data->sftp, dirname->val, dirname->len, mode));
+	RETURN_BOOL(!libssh2_sftp_mkdir_ex(data->sftp, filename, filename_len, mode));
 }
 /* }}} */
 
-/* {{{ proto bool ssh2_sftp_rmdir(resource sftp, string dirname)
+/* {{{ proto bool ssh2_sftp_rmdir(resource sftp, string filename)
  */
 PHP_FUNCTION(ssh2_sftp_rmdir)
 {
 	php_ssh2_sftp_data *data;
 	zval *zsftp;
-	zend_string *dirname;
+	char *filename;
+	int filename_len;
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), "rS", &zsftp, &dirname) == FAILURE) {
+	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rs", &zsftp, &filename, &filename_len) == FAILURE) {
 		return;
 	}
 
-	if ((data = (php_ssh2_sftp_data *)zend_fetch_resource(Z_RES_P(zsftp), PHP_SSH2_SFTP_RES_NAME, le_ssh2_sftp)) == NULL) {
-		RETURN_FALSE;
-	}
+	ZEND_FETCH_RESOURCE(data, php_ssh2_sftp_data*, &zsftp, -1, PHP_SSH2_SFTP_RES_NAME, le_ssh2_sftp);
 
-	RETURN_BOOL(!libssh2_sftp_rmdir_ex(data->sftp, dirname->val, dirname->len));
+	RETURN_BOOL(!libssh2_sftp_rmdir_ex(data->sftp, filename, filename_len));
 }
 /* }}} */
 
@@ -697,26 +717,25 @@ PHP_FUNCTION(ssh2_sftp_chmod)
 {
 	php_ssh2_sftp_data *data;
 	zval *zsftp;
-	zend_string *filename;
-	zend_long mode;
+	char *filename;
+	int filename_len;
+	long mode;
 	LIBSSH2_SFTP_ATTRIBUTES attrs;
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), "rSl", &zsftp, &filename, &mode) == FAILURE) {
+	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rsl", &zsftp, &filename, &filename_len, &mode) == FAILURE) {
 		return;
 	}
 
-	if (ZSTR_LEN(filename) < 1) {
+	if (filename_len < 1) {
 		RETURN_FALSE;
 	}
 
-	if ((data = (php_ssh2_sftp_data *)zend_fetch_resource(Z_RES_P(zsftp), PHP_SSH2_SFTP_RES_NAME, le_ssh2_sftp)) == NULL) {
-		RETURN_FALSE;
-	}
+	ZEND_FETCH_RESOURCE(data, php_ssh2_sftp_data*, &zsftp, -1, PHP_SSH2_SFTP_RES_NAME, le_ssh2_sftp);
 
 	attrs.permissions = mode;
 	attrs.flags = LIBSSH2_SFTP_ATTR_PERMISSIONS;
 
-	RETURN_BOOL(!libssh2_sftp_stat_ex(data->sftp, filename->val, filename->len, LIBSSH2_SFTP_SETSTAT, &attrs));
+	RETURN_BOOL(!libssh2_sftp_stat_ex(data->sftp, filename, filename_len, LIBSSH2_SFTP_SETSTAT, &attrs));
 }
 /* }}} */
 
@@ -730,18 +749,17 @@ static void php_ssh2_sftp_stat_func(INTERNAL_FUNCTION_PARAMETERS, int stat_type)
 	php_ssh2_sftp_data *data;
 	LIBSSH2_SFTP_ATTRIBUTES attrs;
 	zval *zsftp;
-	zend_string *path;
+	char *path;
+	int path_len;
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), "rS", &zsftp, &path) == FAILURE) {
+	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rs", &zsftp, &path, &path_len) == FAILURE) {
 		return;
 	}
 
-	if ((data = (php_ssh2_sftp_data *)zend_fetch_resource(Z_RES_P(zsftp), PHP_SSH2_SFTP_RES_NAME, le_ssh2_sftp)) == NULL) {
-		RETURN_FALSE;
-	}
+	ZEND_FETCH_RESOURCE(data, php_ssh2_sftp_data*, &zsftp, -1, PHP_SSH2_SFTP_RES_NAME, le_ssh2_sftp);
 
-	if (libssh2_sftp_stat_ex(data->sftp, path->val, path->len, stat_type, &attrs)) {
-		php_error_docref(NULL, E_WARNING, "Failed to stat remote file");
+	if (libssh2_sftp_stat_ex(data->sftp, path, path_len, stat_type, &attrs)) {
+		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Failed to stat remote file");
 		RETURN_FALSE;
 	}
 
@@ -794,17 +812,16 @@ PHP_FUNCTION(ssh2_sftp_symlink)
 {
 	php_ssh2_sftp_data *data;
 	zval *zsftp;
-	zend_string *targ, *link;
+	char *targ, *link;
+	int targ_len, link_len;
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), "rSS", &zsftp, &targ, &link) == FAILURE) {
+	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rss", &zsftp, &targ, &targ_len, &link, &link_len) == FAILURE) {
 		return;
 	}
 
-	if ((data = (php_ssh2_sftp_data *)zend_fetch_resource(Z_RES_P(zsftp), PHP_SSH2_SFTP_RES_NAME, le_ssh2_sftp)) == NULL) {
-		RETURN_FALSE;
-	}
+	ZEND_FETCH_RESOURCE(data, php_ssh2_sftp_data*, &zsftp, -1, PHP_SSH2_SFTP_RES_NAME, le_ssh2_sftp);
 
-	RETURN_BOOL(!libssh2_sftp_symlink_ex(data->sftp, targ->val, targ->len, link->val, link->len, LIBSSH2_SFTP_SYMLINK));
+	RETURN_BOOL(!libssh2_sftp_symlink_ex(data->sftp, targ, targ_len, link, link_len, LIBSSH2_SFTP_SYMLINK));
 }
 /* }}} */
 
@@ -814,24 +831,22 @@ PHP_FUNCTION(ssh2_sftp_readlink)
 {
 	php_ssh2_sftp_data *data;
 	zval *zsftp;
-	zend_string *link;
-	int targ_len = 0;
+	char *link;
+	int targ_len = 0, link_len;
 	char targ[8192];
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), "rS", &zsftp, &link) == FAILURE) {
+	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rs", &zsftp, &link, &link_len) == FAILURE) {
 		return;
 	}
 
-	if ((data = (php_ssh2_sftp_data *)zend_fetch_resource(Z_RES_P(zsftp), PHP_SSH2_SFTP_RES_NAME, le_ssh2_sftp)) == NULL) {
+	ZEND_FETCH_RESOURCE(data, php_ssh2_sftp_data*, &zsftp, -1, PHP_SSH2_SFTP_RES_NAME, le_ssh2_sftp);
+
+	if ((targ_len = libssh2_sftp_symlink_ex(data->sftp, link, link_len, targ, 8192, LIBSSH2_SFTP_READLINK)) < 0) {
+		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Unable to read link '%s'", link);
 		RETURN_FALSE;
 	}
 
-	if ((targ_len = libssh2_sftp_symlink_ex(data->sftp, link->val, link->len, targ, 8192, LIBSSH2_SFTP_READLINK)) < 0) {
-		php_error_docref(NULL, E_WARNING, "Unable to read link '%s'", ZSTR_VAL(link));
-		RETURN_FALSE;
-	}
-
-	RETURN_STRINGL(targ, targ_len);
+	RETURN_STRINGL(targ, targ_len, 1);
 }
 /* }}} */
 
@@ -841,24 +856,22 @@ PHP_FUNCTION(ssh2_sftp_realpath)
 {
 	php_ssh2_sftp_data *data;
 	zval *zsftp;
-	zend_string *link;
-	int targ_len = 0;
+	char *link;
+	int targ_len = 0, link_len;
 	char targ[8192];
 
-	if (zend_parse_parameters(ZEND_NUM_ARGS(), "rS", &zsftp, &link) == FAILURE) {
+	if (zend_parse_parameters(ZEND_NUM_ARGS() TSRMLS_CC, "rs", &zsftp, &link, &link_len) == FAILURE) {
 		return;
 	}
 
-	if ((data = (php_ssh2_sftp_data *)zend_fetch_resource(Z_RES_P(zsftp), PHP_SSH2_SFTP_RES_NAME, le_ssh2_sftp)) == NULL) {
+	ZEND_FETCH_RESOURCE(data, php_ssh2_sftp_data*, &zsftp, -1, PHP_SSH2_SFTP_RES_NAME, le_ssh2_sftp);
+
+	if ((targ_len = libssh2_sftp_symlink_ex(data->sftp, link, link_len, targ, 8192, LIBSSH2_SFTP_REALPATH)) < 0) {
+		php_error_docref(NULL TSRMLS_CC, E_WARNING, "Unable to resolve realpath for '%s'", link);
 		RETURN_FALSE;
 	}
 
-	if ((targ_len = libssh2_sftp_symlink_ex(data->sftp, link->val, link->len, targ, 8192, LIBSSH2_SFTP_REALPATH)) < 0) {
-		php_error_docref(NULL, E_WARNING, "Unable to resolve realpath for '%s'", link->val);
-		RETURN_FALSE;
-	}
-
-	RETURN_STRINGL(targ, targ_len);
+	RETURN_STRINGL(targ, targ_len, 1);
 }
 /* }}} */
 
